@@ -7,8 +7,7 @@
 #define SOUP_SESSION_PRIVATE_H 1
 
 #include "soup-session.h"
-#include "soup-connection.h"
-#include "soup-message-queue.h"
+#include "soup-message-private.h"
 #include "soup-proxy-uri-resolver.h"
 
 G_BEGIN_DECLS
@@ -17,19 +16,20 @@ G_BEGIN_DECLS
 SoupMessageQueue     *soup_session_get_queue            (SoupSession          *session);
 
 SoupMessageQueueItem *soup_session_make_connect_message (SoupSession          *session,
-							 SoupAddress          *server_addr);
-SoupConnection       *soup_session_get_connection       (SoupSession          *session,
+							 SoupConnection       *conn);
+gboolean              soup_session_get_connection       (SoupSession          *session,
 							 SoupMessageQueueItem *item,
 							 gboolean             *try_pruning);
 gboolean              soup_session_cleanup_connections  (SoupSession          *session,
 							 gboolean              prune_idle);
-void                  soup_session_connection_failed    (SoupSession          *session,
-							 SoupConnection       *conn,
-							 guint                 status);
-
 void                  soup_session_send_queue_item      (SoupSession          *session,
 							 SoupMessageQueueItem *item,
-							 SoupConnection       *conn);
+							 SoupMessageCompletionFn completion_cb);
+void                  soup_session_unqueue_item         (SoupSession          *session,
+							 SoupMessageQueueItem *item);
+void                  soup_session_set_item_status      (SoupSession          *session,
+							 SoupMessageQueueItem *item,
+							 guint                 status_code);
 
 G_END_DECLS
 
